#!/usr/bin/python3 -i
#
# Copyright (c) 2013-2024, The Khronos Group Inc.
#
# SPDX-License-Identifier: Apache-2.0

from collections import OrderedDict, namedtuple
from enum import Enum
from functools import reduce
from pathlib import Path

from generator import OutputGenerator, write
from spec_tools.attributes import (ExternSyncEntry, LengthEntry,
                                   has_any_optional_in_param,
                                   parse_optional_from_param)
from spec_tools.conventions import ProseListFormats as plf
from spec_tools.data_structures import DictOfStringSets
from spec_tools.util import (findNamedElem, findTypedElem, getElemName,
                             getElemType)
from spec_tools.validity import ValidityCollection, ValidityEntry


class UnhandledCaseError(NotImplementedError):
    def __init__(self, msg=None):
        if msg:
            super().__init__('Got a case in the validity generator that we have not explicitly handled: ' + msg)
        else:
            super().__init__('Got a case in the validity generator that we have not explicitly handled.')


def _genericIterateIntersection(a, b):
    """Iterate through all elements in a that are also in b.

    Somewhat like a set's intersection(),
    but not type-specific so it can work with OrderedDicts, etc.
    It also returns a generator instead of a set,
    so you can pick what container type you'd like,
    if any.
    """
    return (x for x in a if x in b)


def _make_ordered_dict(gen):
    """Make an ordered dict (with None as the values) from a generator."""
    return OrderedDict(((x, None) for x in gen))


def _orderedDictIntersection(a, b):
    return _make_ordered_dict(_genericIterateIntersection(a, b))


def _genericIsDisjoint(a, b):
    """Return true if nothing in a is also in b.

    Like a set's is_disjoint(),
    but not type-specific so it can work with OrderedDicts, etc.
    """
    for _ in _genericIterateIntersection(a, b):
        return False
    # if we never enter the loop...
    return True


_WCHAR = "wchar_t"
_CHAR = "char"
_CHARACTER_TYPES = {_CHAR, _WCHAR}


class StateRelationship(Enum):
    # For commands that begin a state
    BEGIN = 0
    # For commands that end a state
    END = 1
    # For commands that require being in a state without changing state
    CHECK = 2


STATE_ATTRIBUTES = {
    StateRelationship.BEGIN: 'beginvalidstate',
    StateRelationship.END: 'endvalidstate',
    StateRelationship.CHECK: 'checkvalidstate',
}


class StateCollection:
    """Stores associations between states and commands."""

    def __init__(self):
        # Mapping from states to commands
        self.states_to_commands = [DictOfStringSets(), DictOfStringSets(), DictOfStringSets()]

        # Mapping from commands to states
        self.commands_to_states = [DictOfStringSets(), DictOfStringSets(), DictOfStringSets()]

    def get_commands(self, state_name, relationship: StateRelationship):
        return self.states_to_commands[relationship.value].get(state_name)

    def get_states(self, command_name, relationship: StateRelationship):
        return self.commands_to_states[relationship.value].get(command_name)

    def add_command(self, command_name, state_name, relationship: StateRelationship):
        self.states_to_commands[relationship.value].add(state_name, command_name)
        self.commands_to_states[relationship.value].add(command_name, state_name)


class ValidityOutputGenerator(OutputGenerator):
    """ValidityOutputGenerator - subclass of OutputGenerator.

    Generates AsciiDoc includes of valid usage information, for reference
    pages and the specification. Similar to DocOutputGenerator.

    ---- methods ----
    ValidityOutputGenerator(errFile, warnFile, diagFile) - args as for
    OutputGenerator. Defines additional internal state.
    ---- methods overriding base class ----
    beginFile(genOpts)
    endFile()
    beginFeature(interface, emit)
    endFeature()
    genCmd(cmdinfo)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.currentExtension = ''

    @property
    def null(self):
        """Preferred spelling of NULL.

        Delegates to the object implementing ConventionsBase.
        """
        return self.conventions.null

    @property
    def structtype_member_name(self):
        """Return name of the structure type member.

        Delegates to the object implementing ConventionsBase.
        """
        return self.conventions.structtype_member_name

    @property
    def nextpointer_member_name(self):
        """Return name of the structure pointer chain member.

        Delegates to the object implementing ConventionsBase.
        """
        return self.conventions.nextpointer_member_name

    def makeProseList(self, elements, fmt=plf.AND,
                      comma_for_two_elts=False, *args, **kwargs):
        """Make a (comma-separated) list for use in prose.

        Adds a connective (by default, 'and')
        before the last element if there are more than 1.

        Optionally adds a quantifier (like 'any') before a list of 2 or more,
        if specified by fmt.

        Delegates to the object implementing ConventionsBase.
        """
        if not elements:
            raise ValueError(
                'Cannot pass an empty list if you are trying to make a prose list.')
        return self.conventions.makeProseList(elements,
                                              fmt,
                                              with_verb=False,
                                              comma_for_two_elts=comma_for_two_elts,
                                              *args, **kwargs)

    def makeProseListIs(self, elements, fmt=plf.AND,
                        comma_for_two_elts=False, *args, **kwargs):
        """Make a (comma-separated) list for use in prose, followed by either 'is' or 'are' as appropriate.

        Adds a connective (by default, 'and')
        before the last element if there are more than 1.

        Optionally adds a quantifier (like 'any') before a list of 2 or more,
        if specified by fmt.

        Delegates to the object implementing ConventionsBase.
        """
        if not elements:
            raise ValueError(
                'Cannot pass an empty list if you are trying to make a prose list.')
        return self.conventions.makeProseList(elements,
                                              fmt,
                                              with_verb=True,
                                              comma_for_two_elts=comma_for_two_elts,
                                              *args, **kwargs)

    def makeValidityCollection(self, entity_name):
        """Create a ValidityCollection object, passing along our Conventions."""
        return ValidityCollection(entity_name, self.conventions)

    def populateStructChildren(self):
        """Populate self.struct_children."""
        self.struct_children = DictOfStringSets()
        d = self.struct_children
        for name, t in self.registry.typedict.items():
            d.add_key(name)
            parentstruct = t.elem.get('parentstruct')
            if parentstruct is not None:
                d.add(parentstruct, name)

    def populateStates(self):
        """Populate self.states."""
        # OpenXR-specific
        self.states = StateCollection()
        for name, cmdinfo in self.registry.cmddict.items():
            if not cmdinfo.required:
                continue
            self.recordStateRelationship(cmdinfo.elem, name, StateRelationship.BEGIN)
            self.recordStateRelationship(cmdinfo.elem, name, StateRelationship.END)
            self.recordStateRelationship(cmdinfo.elem, name, StateRelationship.CHECK)

    def beginFile(self, genOpts):
        if not genOpts.conventions:
            raise RuntimeError(
                'Must specify conventions object to generator options')
        self.conventions = genOpts.conventions
        # Vulkan says 'must: be a valid pointer' a lot, OpenXR just says
        # 'must: be a pointer'.
        self.valid_pointer_text = ' '.join(
            (x for x in (self.conventions.valid_pointer_prefix, 'pointer') if x))

        # OpenXR-only
        self.populateStructChildren()
        self.populateStates()

        OutputGenerator.beginFile(self, genOpts)

    def generateStateValidity(self, validity, command_name):
        begins_states = self.states.get_states(command_name, StateRelationship.BEGIN)
        if begins_states:
            for state_name in sorted(begins_states):
                end_commands = self.states.get_commands(state_name, StateRelationship.END)
                assert end_commands

                entry = ValidityEntry(anchor=(state_name, 'beginstate'))
                entry += 'flink:{} must: not be called more than once without first successfully calling '.format(
                    command_name)
                entry += self.makeProseList(sorted(end_commands), fmt=plf.OR)
                validity += entry

        ends_states = self.states.get_states(command_name, StateRelationship.END)
        if ends_states:
            for state_name in sorted(ends_states):
                begin_commands = self.states.get_commands(state_name, StateRelationship.BEGIN)
                assert begin_commands

                entry = ValidityEntry(anchor=(state_name, 'beginstate'))
                entry += 'flink:{} must: only be called after a successful call to '.format(
                    command_name)
                entry += self.makeProseList(sorted(begin_commands), fmt=plf.OR)
                validity += entry

        checks_states = self.states.get_states(command_name, StateRelationship.CHECK)
        if checks_states:
            for state_name in sorted(checks_states):
                begin_commands = self.states.get_commands(state_name, StateRelationship.BEGIN)
                end_commands = self.states.get_commands(state_name, StateRelationship.END)
                assert begin_commands
                assert end_commands

                begin_command_list = '/'.join('flink:{}'.format(command)
                                              for command in begin_commands)
                end_command_list = '/'.join('flink:{}'.format(command)
                                            for command in end_commands)

                entry = ValidityEntry(anchor=(state_name, 'checkstate'))
                entry += 'flink:{} must: only be called between successful calls to '.format(
                    command_name)
                entry += begin_command_list
                entry += ' and '
                entry += end_command_list
                validity += entry

    def endFile(self):
        OutputGenerator.endFile(self)

    def beginFeature(self, interface, emit):
        # Start processing in superclass
        OutputGenerator.beginFeature(self, interface, emit)
        self.currentExtension = interface.get('name')

    def endFeature(self):
        # Finish processing in superclass
        OutputGenerator.endFeature(self)

    @property
    def struct_macro(self):
        """Get the appropriate format macro for a structure."""
        # delegate to conventions
        return self.conventions.struct_macro

    def makeStructName(self, name):
        """Prepend the appropriate format macro for a structure to a structure type name."""
        # delegate to conventions
        return self.conventions.makeStructName(name)

    def makeParameterName(self, name):
        """Prepend the appropriate format macro for a parameter/member to a parameter name."""
        return 'pname:' + name

    def makeBaseTypeName(self, name):
        """Prepend the appropriate format macro for a 'base type' to a type name."""
        return 'basetype:' + name

    def makeEnumerationName(self, name):
        """Prepend the appropriate format macro for an enumeration type to a enum type name."""
        return 'elink:' + name

    def makeFuncPointerName(self, name):
        """Prepend the appropriate format macro for a function pointer type to a type name."""
        return 'tlink:' + name

    def makeExternalTypeName(self, name):
        """Prepend the appropriate format macro for an external type like uint32_t to a type name."""
        # delegate to conventions
        return self.conventions.makeExternalTypeName(name)

    def makeEnumerantName(self, name):
        """Prepend the appropriate format macro for an enumerate (value) to a enum value name."""
        return 'ename:' + name

    def writeInclude(self, directory, basename, validity: ValidityCollection,
                     threadsafety, commandpropertiesentry=None,
                     successcodes=None, errorcodes=None):
        """Generate an include file.

        directory - subdirectory to put file in (absolute or relative pathname)
        basename - base name of the file
        validity - ValidityCollection to write.
        threadsafety - List (may be empty) of thread safety statements to write.
        successcodes - Optional success codes to document.
        errorcodes - Optional error codes to document.
        """
        # Create subdirectory, if needed
        directory = Path(directory)
        if not directory.is_absolute():
            directory = Path(self.genOpts.directory) / directory
        self.makeDir(directory)

        # Create validity file
        filename = directory / (basename + '.txt')
        self.logMsg('diag', '# Generating include file:', str(filename))

        with open(filename, 'w', encoding='utf-8') as fp:
            write(self.conventions.warning_comment, file=fp)

            # Valid Usage
            if validity:
                write('.Valid Usage (Implicit)', file=fp)
                write('****', file=fp)
                write(validity, file=fp, end='')
                write('****', file=fp)
                write('', file=fp)

            # Host Synchronization
            if threadsafety:
                # The heading of this block differs between projects, so an Asciidoc attribute is used.
                write('.{externsynctitle}', file=fp)
                write('****', file=fp)
                write(threadsafety, file=fp, end='')
                write('****', file=fp)
                write('', file=fp)

            # Command Properties - contained within a block, to avoid table numbering
            if commandpropertiesentry:
                write('.Command Properties', file=fp)
                write('****', file=fp)
                write('[options="header", width="100%"]', file=fp)
                write('|====', file=fp)
                write('|<<VkCommandBufferLevel,Command Buffer Levels>>|<<vkCmdBeginRenderPass,Render Pass Scope>>|<<VkQueueFlagBits,Supported Queue Types>>|<<synchronization-pipeline-stages-types,Pipeline Type>>', file=fp)
                write(commandpropertiesentry, file=fp)
                write('|====', file=fp)
                write('****', file=fp)
                write('', file=fp)

            # Success Codes - contained within a block, to avoid table numbering
            if successcodes or errorcodes:
                write('.Return Codes', file=fp)
                write('****', file=fp)
                if successcodes:
                    write('ifndef::doctype-manpage[]', file=fp)
                    write('<<fundamentals-successcodes,Success>>::', file=fp)
                    write('endif::doctype-manpage[]', file=fp)
                    write('ifdef::doctype-manpage[]', file=fp)
                    write('On success, this command returns::', file=fp)
                    write('endif::doctype-manpage[]', file=fp)
                    write(successcodes, file=fp)
                if errorcodes:
                    write('ifndef::doctype-manpage[]', file=fp)
                    write('<<fundamentals-errorcodes,Failure>>::', file=fp)
                    write('endif::doctype-manpage[]', file=fp)
                    write('ifdef::doctype-manpage[]', file=fp)
                    write('On failure, this command returns::', file=fp)
                    write('endif::doctype-manpage[]', file=fp)
                    write(errorcodes, file=fp)
                write('****', file=fp)
                write('', file=fp)


    def paramIsStaticArray(self, param):
        """Check if the parameter passed in is a static array."""
        tail = param.find('name').tail
        return tail and tail[0] == '['

    def paramIsConst(self, param):
        """Check if the parameter passed in has a type that mentions const."""
        return param.text is not None and 'const' in param.text

    def staticArrayLength(self, param):
        """Get the length of a parameter that's been identified as a static array."""
        paramenumsize = param.find('enum')
        if paramenumsize is not None:
            return self.makeEnumerantName(paramenumsize.text)

        return param.find('name').tail[1:-1]


    def getHandleDispatchableAncestors(self, typename):
        """Get the ancestors of a handle object."""
        ancestors = []
        current = typename
        while True:
            current = self.getHandleParent(current)
            if current is None:
                return ancestors
            if self.isHandleTypeDispatchable(current):
                ancestors.append(current)

    def isHandleTypeDispatchable(self, handlename):
        """Check if a parent object is dispatchable or not."""
        handle = self.registry.tree.find(
            "types/type/[name='" + handlename + "'][@category='handle']")
        if handle is not None and getElemType(handle) == 'VK_DEFINE_HANDLE':
            return True
        else:
            return False

    def isHandleOptional(self, param, params):
        # Simple, if it is optional, return true
        if has_any_optional_in_param(param):
            return True

        # If no validity is being generated, it usually means that validity is complex and not absolute, so let's say yes.
        if param.get('noautovalidity') is not None:
            return True

        # If the parameter is an array and we have not already returned, find out if any of the len parameters are optional
        if self.paramIsArray(param):
            for length in LengthEntry.parse_len_from_param(param):
                if not length.other_param_name:
                    # do not care about constants or "null-terminated"
                    continue

                other_param = findNamedElem(params, length.other_param_name)
                if other_param is None:
                    self.logMsg('warn', length.other_param_name,
                                'is listed as a length for parameter', param, 'but no such parameter exists')
                if other_param and has_any_optional_in_param(other_param):
                    return True

        return False

    def makeParamValidityPre(self, param, params):
        """Make the start of an entry for a parameter's validity, including a chunk of text if it is an array."""
        param_name = getElemName(param)
        paramtype = getElemType(param)

        # General pre-amble. Check optionality and add stuff.
        entry = ValidityEntry(anchor=(param_name, 'parameter'))

        if self.paramIsStaticArray(param):
            if paramtype not in _CHARACTER_TYPES:
                entry += 'Any given element of '
            return entry

        optional = parse_optional_from_param(param)
        is_optional = optional[0]

        if self.paramIsArray(param) and param.get('len') != LengthEntry.NULL_TERMINATED_STRING:
            # Find all the parameters that are called out as optional,
            # so we can document that they might be zero, and the array may be ignored
            optionallengths = []
            for length in LengthEntry.parse_len_from_param(param):
                if not length.other_param_name:
                    # Only care about length entries that are parameter names
                    continue

                other_param = findNamedElem(params, length.other_param_name)
                other_param_optional = has_any_optional_in_param(other_param)

                if other_param is None or not other_param_optional:
                    # Do not care about not-found params or non-optional params
                    continue

                if self.paramIsPointer(other_param):
                    optionallengths.append(
                        'the value referenced by ' + self.makeParameterName(length.other_param_name))
                else:
                    optionallengths.append(
                        self.makeParameterName(length.other_param_name))

            # Document that these arrays may be ignored if any of the length values are 0
            if optionallengths or is_optional:
                entry += 'If '
            if optionallengths:
                entry += self.makeProseListIs(optionallengths, fmt=plf.OR)
                entry += ' not %s, ' % self.conventions.zero
            # TODO enabling this in OpenXR, as used in Vulkan, causes nonsensical things like
            # "If pname:propertyCapacityInput is not `0`, and pname:properties is not `NULL`, pname:properties must: be a pointer to an array of pname:propertyCapacityInput slink:XrApiLayerProperties structures"
            # if optionallengths and is_optional:
            #     entry += 'and '
            # if is_optional:
            #     entry += self.makeParameterName(param_name)
            #     # TODO switch when cosmetic changes OK
            #     # entry += ' is not {}, '.format(self.null)
            #     entry += ' is not `NULL`, '

            return entry

        if any(optional):
            # Do not generate this stub for bitflags
            type_category = self.getTypeCategory(paramtype)
            if type_category != 'bitmask' and is_optional:
                if self.paramIsArray(param) or self.paramIsPointer(param):
                    optional_val = self.null
                elif type_category == 'handle':
                    if self.isHandleTypeDispatchable(paramtype):
                        optional_val = self.null
                    else:
                        optional_val = 'dlink:' + self.conventions.api_prefix + 'NULL_HANDLE'
                else:
                    optional_val = self.conventions.zero

                entry += 'If {} is not {}, '.format(
                    self.makeParameterName(param_name),
                    optional_val)
            return entry

        # If none of the early returns happened, we at least return an empty
        # entry with an anchor.
        return entry

    def createValidationLineForParameterImpl(self, blockname, param, params, typetext, see_also):
        """Make the generic validity portion used for all parameters.

        May return None if nothing to validate.
        """
        if param.get('noautovalidity') is not None:
            return None

        validity = self.makeValidityCollection(blockname)
        param_name = getElemName(param)
        paramtype = getElemType(param)

        entry = self.makeParamValidityPre(param, params)
        entry += '{} must: be '.format(self.makeParameterName(param_name))

        def add_see_also(entry):
            if see_also:
                entry += '. '
                entry += see_also

        optional = parse_optional_from_param(param)
        if self.paramIsStaticArray(param) and paramtype in _CHARACTER_TYPES:
            # TODO this is a minor hack to determine if this is a command parameter or a struct member
            if self.paramIsConst(param) or blockname.startswith(self.conventions.type_prefix):
                if paramtype != _CHAR:
                    raise UnhandledCaseError("input arrays of wchar_t are not yet handled")
                entry += 'a null-terminated UTF-8 string whose length is less than or equal to '
                entry += self.staticArrayLength(param)
            else:
                # This is a command's output parameter
                entry += 'a '
                if paramtype == _WCHAR:
                    entry += "wide "
                entry += 'character array of length %s ' % self.staticArrayLength(param)
            add_see_also(entry)
            validity += entry
            return validity

        elif self.paramIsArray(param):
            # Arrays. These are hard to get right, apparently

            lengths = LengthEntry.parse_len_from_param(param)
            if lengths is None:
                raise RuntimeError("Logic error: decided this was an array but there is no len attribute")

            for i, length in enumerate(lengths):
                if i == 0:
                    # If the first index, make it singular.
                    entry += 'a '
                    array_text = 'an array'
                    pointer_text = self.valid_pointer_text
                else:
                    array_text = 'arrays'
                    pointer_text = self.valid_pointer_text + 's'

                if length.null_terminated:
                    # This should always be the last thing.
                    # If it ever is not for some bizarre reason, then this will need some massaging.
                    entry += 'null-terminated '
                elif length.number == 1:
                    entry += pointer_text
                    entry += ' to '
                else:
                    entry += pointer_text
                    entry += ' to '
                    entry += array_text
                    entry += ' of '
                    # Handle equations, which are currently denoted with latex
                    if length.math:
                        # Handle equations, which are currently denoted with latex
                        entry += str(length)
                    else:
                        entry += self.makeParameterName(str(length))
                    entry += ' '

            # Void pointers do not actually point at anything - remove the word "to"
            if paramtype == 'void':
                if lengths[-1].number == 1:
                    if len(lengths) > 1:
                        # Take care of the extra s added by the post array chunk function. #HACK#
                        entry.drop_end(5)
                    else:
                        entry.drop_end(4)

                    # This has not been hit, so this has not been tested recently.
                    raise UnhandledCaseError("Got void pointer param/member with last length 1")
                else:
                    # An array of void values is a byte array.
                    entry += 'byte'

            elif paramtype == _CHAR:
                # A null terminated array of chars is a string
                if lengths[-1].null_terminated:
                    entry += 'UTF-8 string'
                else:
                    # Else it is just a bunch of chars
                    entry += 'char value'

            elif self.paramIsConst(param):
                # If a value is "const" that means it will not get modified, so it must be valid going into the function.
                if 'const' in param.text:

                    if not self.isStructAlwaysValid(paramtype):
                        entry += 'valid '

            # Check if the array elements are optional
            array_element_optional = len(optional) == len(lengths) + 1 \
                and optional[-1]
            if array_element_optional and self.getTypeCategory(paramtype) != 'bitmask':  # bitmask is handled later
                entry += 'or dlink:' + self.conventions.api_prefix + 'NULL_HANDLE '

            entry += typetext

            # pluralize
            if len(lengths) > 1 or (lengths[0] != 1 and not lengths[0].null_terminated):
                entry += 's'
            add_see_also(entry)

            return self.handleRequiredBitmask(blockname, param, paramtype, entry)

        if self.paramIsPointer(param):
            # Handle pointers - which are really special case arrays (i.e. they do not have a length)
            # TODO  should do something here if someone ever uses some intricate comma-separated `optional`
            pointercount = param.find('type').tail.count('*')
            # Treat void* as an int
            if paramtype == 'void':
                # If there is only void*, it is just optional int - we do not need any language.
                if pointercount == 1 and optional[0]:
                    return None  # early return
                # Treat the inner-most void* as an int
                pointercount -= 1

            # Could be multi-level pointers (e.g. ppData - pointer to a pointer). Handle that.
            entry += 'a '
            entry += (self.valid_pointer_text + ' to a ') * pointercount

            # Handle void* and pointers to it
            if paramtype == 'void':
                if not optional[pointercount]:
                    typetext = 'pointer value'
                else:
                    # The old conditional added "pointer value" if optional was not specified (which implies "false"),
                    # or if optional[pointercount] was truthy, which just means non-empty string.
                    # old comment:
                    # The last void* is just optional int (e.g. to be filled by the impl.)
                    raise UnhandledCaseError("Conditional previously was mixed")

            # If a value is "const" that means it will not get modified, so it must be valid going into the function.
            elif self.paramIsConst(param) and paramtype != 'void':
                entry += 'valid '

            entry += typetext
            add_see_also(entry)
            return self.handleRequiredBitmask(blockname, param, paramtype, entry)

        # Add additional line for non-optional bitmasks
        if self.getTypeCategory(paramtype) == 'bitmask':
            # TODO does not really handle if someone tries something like optional="true,false"
            # TODO OpenXR has 0 or a valid combination of flags, for optional things.
            # Vulkan does not...
            isMandatory = not optional[0]
            if not isMandatory:
                entry += self.conventions.zero
                entry += ' or '
            # Non-pointer, non-optional things must be valid
            entry += 'a valid {}'.format(typetext)

            return self.handleRequiredBitmask(blockname, param, paramtype, entry)

        # Non-pointer, non-optional things must be valid
        entry += 'a valid {}'.format(typetext)
        add_see_also(entry)
        return entry

    def handleRequiredBitmask(self, blockname, param, paramtype, entry):
        # TODO does not really handle if someone tries something like optional="true,false"
        optional = parse_optional_from_param(param)
        if self.getTypeCategory(paramtype) != 'bitmask' or \
                any(optional):
            return entry
        if self.paramIsPointer(param) and not self.paramIsArray(param):
            # This is presumably an output parameter
            return entry

        param_name = getElemName(param)
        # If mandatory, then we need two entries instead of just one.
        validity = self.makeValidityCollection(blockname)
        validity += entry

        entry2 = ValidityEntry(anchor=(param_name, 'requiredbitmask'))
        if self.paramIsArray(param):
            entry2 += 'Each element of '
        entry2 += '{} must: not be {}'.format(
            self.makeParameterName(param_name), self.conventions.zero)
        validity += entry2
        return validity

    def isBaseHeaderType(self, typename):
        """Returns true if the type is a struct that is a "base header" type."""
        assert self.registry
        info = self.registry.typedict.get(typename)
        if not info:
            return False
        members = info.getMembers()
        type_member = findNamedElem(members, self.structtype_member_name)
        if type_member is None:
            return False

        # If we have a type member without specified values, it is a base header.
        return type_member.get('values') is None

    def createValidationLineForParameter(self, blockname, param, params, typecategory):
        """Make an entire validation entry for a given parameter."""
        assert self.registry
        param_name = getElemName(param)
        paramtype = getElemType(param)
        see_also = None
        is_array = self.paramIsArray(param)
        is_pointer = self.paramIsPointer(param)
        needs_recursive_validity = (is_array
                                    or is_pointer
                                    or not self.isStructAlwaysValid(paramtype))
        typetext = None
        if paramtype in ('void', _CHAR):
            # Chars and void are special cases - we call the impl function,
            # but do not use the typetext.
            # A null-terminated char array is a string, else it is chars.
            # An array of void values is a byte array, a void pointer is just a pointer to nothing in particular
            typetext = ''

        elif typecategory == 'bitmask':
            bitsname = paramtype.replace('Flags', 'FlagBits')
            bitselem = self.registry.tree.find("enums[@name='" + bitsname + "']")
            if bitselem is None or len(bitselem.findall('enum')) == 0:
                # Empty bit mask: presumably just a placeholder (or only in
                # an extension not enabled for this build)

                entry = ValidityEntry(
                    anchor=(param_name, 'zerobitmask'))
                entry += self.makeParameterName(param_name)
                entry += ' must: be '
                entry += self.conventions.zero
                # Early return
                return entry

            is_const = self.paramIsConst(param)

            if is_array:
                if is_const:
                    # input an array of bitmask values
                    template = 'combinations of {bitsname} value'
                else:
                    template = '{paramtype} value'
            elif is_pointer:
                if is_const:
                    template = 'combination of {bitsname} values'
                else:
                    template = '{paramtype} value'
            else:
                template = 'combination of {bitsname} values'

            # The above few cases all use makeEnumerationName, just with different context.
            typetext = template.format(
                bitsname=self.makeEnumerationName(bitsname),
                paramtype=self.makeEnumerationName(paramtype))

        elif typecategory == 'handle':
            typetext = '{} handle'.format(self.makeStructName(paramtype))

        elif typecategory == 'enum':
            typetext = '{} value'.format(self.makeEnumerationName(paramtype))

        elif typecategory == 'funcpointer':
            typetext = '{} value'.format(self.makeFuncPointerName(paramtype))

        elif typecategory == 'struct':
            if needs_recursive_validity:
                if self.isBaseHeaderType(paramtype):
                    typetext = '{}-based structure'.format(
                        self.makeStructName(paramtype))
                    child_structs = self.keepOnlyRequired(self.struct_children.get(paramtype, []),
                                                          self.registry.typedict)
                    if child_structs:
                        see_also = 'See also: {}'.format(', '.join(self.makeStructName(t) for t in sorted(child_structs)))
                else:
                    typetext = '{} structure'.format(
                        self.makeStructName(paramtype))

        elif typecategory == 'union':
            if needs_recursive_validity:
                typetext = '{} union'.format(self.makeStructName(paramtype))

        elif self.paramIsArray(param) or self.paramIsPointer(param):
            if typecategory == 'basetype':
                typetext = '{} value'.format(self.makeBaseTypeName(paramtype))
            elif typecategory is None:
                typetext = '{} value'.format(self.makeExternalTypeName(paramtype))
            else:
                raise UnhandledCaseError()

        elif typecategory is None:
            if not self.isStructAlwaysValid(paramtype):
                typetext = '{} value'.format(
                    self.makeExternalTypeName(paramtype))

            # "a valid uint32_t value" does not make much sense.
            pass

        # If any of the above conditions matched and set typetext,
        # we call using it.
        if typetext is not None:
            return self.createValidationLineForParameterImpl(
                blockname, param, params, typetext, see_also)
        return None

    def makeHandleValidityParent(self, param, params):
        """Make a validity entry for a handle's parent object.

        Creates 'parent' VUID.
        """
        param_name = getElemName(param)
        paramtype = getElemType(param)

        # Deal with handle parents
        handleparent = self.getHandleParent(paramtype)
        if handleparent is None:
            return None

        otherparam = findTypedElem(params, handleparent)
        if otherparam is None:
            return None

        parent_name = getElemName(otherparam)
        entry = ValidityEntry(anchor=(param_name, 'parent'))

        is_optional = self.isHandleOptional(param, params)

        if self.paramIsArray(param):
            template = 'Each element of {}'
            if is_optional:
                template += ' that is a valid handle'
        elif is_optional:
            template = 'If {} is a valid handle, it'
        else:
            # not optional, not an array. Just say the parameter name.
            template = '{}'

        entry += template.format(self.makeParameterName(param_name))

        entry += ' must: have been created, allocated, or retrieved from {}'.format(
            self.makeParameterName(parent_name))

        return entry

    def getUnnamedCommonParentOfParams(self, input_handles):
        """Return the nearest common ancestor handle type for all handle parameters given."""
        if not input_handles or len(input_handles) < 2:
            return None

        # Get a list (well, an OrderedDict pretending to be a set)
        # of the full ancestor chain for each input handle.
        ancestor_map = {
            h: _make_ordered_dict(self.iterateHandleAncestors(getElemType(h)))
            for h in input_handles
        }

        # all handle types supplied to this function
        handle_types = set(getElemType(h) for h in input_handles)

        # Only consider the handles without any ancestors also being
        # supplied to this function.
        top_level_handles = set(h for h in input_handles
                                if _genericIsDisjoint(ancestor_map[h], handle_types))

        # Remove all non-top-level-handles from further consideration
        for removal in set(ancestor_map.keys()).difference(top_level_handles):
            del(ancestor_map[removal])

        if len(ancestor_map) <= 1:
            # If only one top level,
            # it has no other top level to share an unnamed parent with.
            return None

        # Find intersection (ordered!) of all ancestor lists.
        #all_ancestor_lists = ancestor_map.values()
        common_ancestors = reduce(_orderedDictIntersection,
                                  ancestor_map.values())

        # The first common ancestor should be the nearest one.
        # We already ensured we have at least one.
        common_ancestor = next(iter(common_ancestors), None)
        return common_ancestor

    def makeHandlesValidityCommonAncestor(self, handles, params):
        """Make an validity entry for a common ancestor between handles, not explicitly passed as a param.

        Only handles parent validity for signatures taking multiple handles
        without any ancestors also being supplied to the function.
        (e.g. "Each of x, y, and z must: come from the same slink:ParentHandle")
        See self.makeHandleValidityParent() for instances where the parent
        handle is named and also passed.

        Creates 'commonparent' VUID.
        """
        if not handles or len(handles) < 2:
            return None

        # Get just the inputs: pointers to handles are outputs,
        # unless they're const (then they're probably an input array).
        # Inputs are the only ones we generate this sort of validity for.

        def is_input_handle(h):
            # NOTE: having paramIsPointer before the "or" messes up formatters...
            return (self.paramIsConst(h)
                    or not self.paramIsPointer(h))

        input_handles = [h
                         for h in handles
                         if is_input_handle(h)]
        common_ancestor = self.getUnnamedCommonParentOfParams(input_handles)
        if common_ancestor is None:
            # No common ancestor to validate
            return None

        parameter_texts = []
        for param in input_handles:
            pname = self.makeParameterName(getElemName(param))
            if self.paramIsArray(param):
                parameter_texts.append('the elements of {}'.format(pname))
            else:
                parameter_texts.append(pname)

        entry = ValidityEntry(anchor=('commonparent',))
        entry += self.makeProseList(sorted(parameter_texts), fmt=plf.EACH_AND)

        if any((self.isHandleOptional(h, params) for h in input_handles)):
            entry += ' that are valid handles'

        entry += ' must: have been created, allocated, or retrieved from the same '
        entry += self.makeStructName(common_ancestor)

        return entry

    def makeStructureTypeFromName(self, structname):
        """Create text for a structure type name, like ename:XR_TYPE_CREATE_INSTANCE_INFO"""
        return self.makeEnumerantName(self.conventions.generate_structure_type_from_name(structname))

    def makeStructureTypeValidity(self, structname):
        """Generate an validity line for the type value of a struct.

        Creates VUID named like the member name.
        """
        info = self.registry.typedict.get(structname)
        assert info is not None

        # If this fails (meaning we have something other than a struct in here),
        # then the caller is wrong:
        # probably passing the wrong value for structname.
        members = info.getMembers()
        assert members

        param = findNamedElem(members, self.structtype_member_name)
        # OpenXR gets some structs without a type field in here, so cannot assert
        # assert param is not None
        if param is None:
            return None

        entry = ValidityEntry(
            anchor=(self.structtype_member_name, self.structtype_member_name))
        entry += self.makeParameterName(self.structtype_member_name)
        entry += ' must: be '

        values = param.get('values')
        if values:
            values = self.findRequiredEnums(values.split(','))

        child_structs = self.keepOnlyRequired(self.struct_children.get(structname, []),
                                              self.registry.typedict)
        if child_structs:
            if values:
                print('The struct: {} has children, it may not have a "values" attribute itself.'.format(structname))
            assert not values
            if len(child_structs) > 1:
                entry += 'one of the following XrStructureType values: '
            entry += ', '.join(self.makeStructureTypeFromName(child)
                               for child in sorted(child_structs))
            return entry

        if values:
            # Extract each enumerant value. They could be validated in the
            # same fashion as validextensionstructs in
            # makeStructureExtensionPointer, although that's not relevant in
            # the current extension struct model.
            entry += self.makeProseList((self.makeEnumerantName(v)
                                         for v in values), 'or')
            return entry

        if 'Base' in structname:
            # This type does not even have any values for its type,
            # and it seems like it might be a base struct that we'd expect to lack its own type,
            # so omit the entire statement
            return None

        self.logMsg('warn', 'No values were marked-up for the structure type member of',
                    structname, 'so making one up!')
        entry += self.makeStructureTypeFromName(structname)

        return entry

    def makeStructureExtensionPointer(self, blockname, param):
        """Generate an validity line for the pointer chain member value of a struct."""
        param_name = getElemName(param)

        if param.get('validextensionstructs') is not None:
            self.logMsg('warn', blockname,
                        'validextensionstructs is deprecated/removed', '\n')

        entry = ValidityEntry(
            anchor=(param_name, self.nextpointer_member_name))
        validextensionstructs = self.registry.validextensionstructs.get(
            blockname)
        extensionstructs = []

        if validextensionstructs is not None:
            # Check each structure name and skip it if not required by the
            # generator. This allows tagging extension structs in the XML
            # that are only included in validity when needed for the spec
            # being targeted.
            for struct in validextensionstructs:
                # Unpleasantly breaks encapsulation. Should be a method in the registry class
                t = self.registry.lookupElementInfo(
                    struct, self.registry.typedict)
                if t is None:
                    self.logMsg('warn', 'makeStructureExtensionPointer: struct', struct,
                                'is in a validextensionstructs= attribute but is not in the registry')
                elif t.required:
                    extensionstructs.append('slink:' + struct)
                else:
                    self.logMsg(
                        'diag', 'makeStructureExtensionPointer: struct', struct, 'IS NOT required')

        link = "link:{uri-next-chain}[next structure in a structure chain]"
        entry += f'{self.makeParameterName(param_name)} must: be {self.null} or a valid pointer to the {link}'
        if not extensionstructs:
            return entry
        entry += '. See also: '

        entry += ', '.join(extensionstructs)
        return entry

    def addSharedStructMemberValidity(self, struct, blockname, param, validity):
        """Generate language to independently validate a parameter, for those validated even in output.

        Return value indicates whether it was handled internally (True) or if it may need more validity (False)."""
        param_name = getElemName(param)
        paramtype = getElemType(param)
        if param.get('noautovalidity') is None:

            if self.conventions.is_structure_type_member(paramtype, param_name):
                validity += self.makeStructureTypeValidity(blockname)
                return True

            if self.conventions.is_nextpointer_member(paramtype, param_name):
                # Vulkan: make the addition of validity here conditional.
                # if struct.get('structextends') is None:
                #     validity += self.makeStructureExtensionPointer(blockname, param)
                validity += self.makeStructureExtensionPointer(blockname, param)
                return True
        return False

    def makeOutputOnlyStructValidity(self, cmd, blockname, params):
        """Generate all the valid usage information for a struct that's entirely output.

        That is, it is only ever filled out by the implementation other than
        the structure type and pointer chain members.
        Thus, we only create validity for the pointer chain member.
        """
        # Start the validity collection for this struct
        validity = self.makeValidityCollection(blockname)

        for param in params:
            self.addSharedStructMemberValidity(
                cmd, blockname, param, validity)

        return validity

    def makeStructOrCommandValidity(self, cmd, blockname, params):
        """Generate all the valid usage information for a given struct or command."""
        validity = self.makeValidityCollection(blockname)
        handles = []
        arraylengths = dict()
        for param in params:
            param_name = getElemName(param)
            paramtype = getElemType(param)

            # Valid usage ID tags (VUID) are generated for various
            # conditions based on the name of the block (structure or
            # command), name of the element (member or parameter), and type
            # of VU statement.

            # Get the type's category
            typecategory = self.getTypeCategory(paramtype)

            if not self.addSharedStructMemberValidity(
                    cmd, blockname, param, validity):
                validity += self.createValidationLineForParameter(
                    blockname, param, params, typecategory)

            # Ensure that any parenting is properly validated, and list that a handle was found
            if typecategory == 'handle':
                handles.append(param)

            # Get the array length for this parameter
            lengths = LengthEntry.parse_len_from_param(param)
            if lengths:
                arraylengths.update({length.other_param_name: length
                                     for length in lengths
                                     if length.other_param_name})

        # Any non-optional arraylengths should specify they must be greater than 0
        array_length_params = ((param, getElemName(param))
                               for param in params
                               if getElemName(param) in arraylengths)

        for param, param_name in array_length_params:
            if has_any_optional_in_param(param):
                continue

            length = arraylengths[param_name]
            full_length = length.full_reference

            # Is this just a name of a param? If false, then it is some kind of qualified name (a member of a param for instance)
            simple_param_reference = (len(length.param_ref_parts) == 1)

            # Get all the array dependencies
            arrays = cmd.findall(
                "param/[@len='{}'][@optional='true']".format(full_length))

            # Get all the optional array dependencies, including those not generating validity for some reason
            optionalarrays = arrays + \
                cmd.findall(
                    "param/[@len='{}'][@noautovalidity='true']".format(full_length))

            entry = ValidityEntry(anchor=(full_length, 'arraylength'))
            # Allow lengths to be arbitrary if all their dependents are optional
            if optionalarrays and len(optionalarrays) == len(arrays):
                entry += 'If '
                optional_array_names = (self.makeParameterName(getElemName(array))
                                        for array in optionalarrays)
                entry += self.makeProseListIs(optional_array_names,
                                              fmt=plf.ANY_OR)

                entry += ' not {}, '.format(self.null)

            if self.paramIsPointer(param):
                entry += 'the value referenced by '

            # Split and re-join here to insert pname: around ::
            entry += length.get_human_readable(make_param_name=self.makeParameterName)
            entry += ' must: be greater than '
            entry += self.conventions.zero
            validity += entry

        # Find the parents of all objects referenced in this command
        for param in handles:
            # Do not detect a parent for return values!
            if not self.paramIsPointer(param) or self.paramIsConst(param):
                validity += self.makeHandleValidityParent(param, params)

        # Find the common ancestor of all objects referenced in this command
        validity += self.makeHandlesValidityCommonAncestor(handles, params)

        return validity

    def makeThreadSafetyBlock(self, cmd, paramtext):
        """Generate thread-safety validity entries for cmd/structure"""
        # See also makeThreadSafetyBlock in validitygenerator.py
        validity = self.makeValidityCollection(getElemName(cmd))

        # This text varies between projects, so an Asciidoctor attribute is used.
        extsync_prefix = "{externsyncprefix} "

        # Find and add any parameters that are thread unsafe
        explicitexternsyncparams = cmd.findall(paramtext + "[@externsync]")
        if explicitexternsyncparams is not None:
            for param in explicitexternsyncparams:
                externsyncattribs = ExternSyncEntry.parse_externsync_from_param(param)
                param_name = getElemName(param)

                for attrib in externsyncattribs:
                    entry = ValidityEntry()
                    entry += extsync_prefix
                    if attrib.entirely_extern_sync:
                        if self.paramIsArray(param):
                            entry += 'each member of '
                        elif self.paramIsPointer(param):
                            entry += 'the object referenced by '

                        entry += self.makeParameterName(param_name)

                        if attrib.children_extern_sync:
                            entry += ', and any child handles,'

                    else:
                        entry += attrib.get_human_readable(make_param_name=self.makeParameterName)
                    entry += ' must: be externally synchronized'
                    validity += entry

        # Vulkan-specific
        # For any vkCmd* functions, the command pool is externally synchronized
        if cmd.find('proto/name') is not None and 'vkCmd' in cmd.find('proto/name').text:
            entry = ValidityEntry()
            entry += extsync_prefix
            entry += 'the sname:VkCommandPool that pname:commandBuffer was allocated from must: be externally synchronized'
            validity += entry

        # Find and add any "implicit" parameters that are thread unsafe
        implicitexternsyncparams = cmd.find('implicitexternsyncparams')
        if implicitexternsyncparams is not None:
            for elem in implicitexternsyncparams:
                entry = ValidityEntry()
                entry += extsync_prefix
                entry += elem.text
                entry += ' must: be externally synchronized'
                validity += entry

        return validity

    def findRequiredEnums(self, enums):
        """Check each enumerant name in the enums list and remove it if not
        required by the generator. This allows specifying success and error
        codes for extensions that are only included in validity when needed
        for the spec being targeted."""
        return self.keepOnlyRequired(enums, self.registry.enumdict)

    def findRequiredCommands(self, commands):
        """Check each command name in the commands list and remove it if not
        required by the generator.

        This will allow some state operations to take place before endFile."""
        return self.keepOnlyRequired(commands, self.registry.cmddict)

    def keepOnlyRequired(self, names, info_dict):
        """Check each element name in the supplied dictionary and remove it if not
        required by the generator.

        This will allow some operations to take place before endFile no matter the order of generation."""
        # TODO Unpleasantly breaks encapsulation. Should be a method in the registry class

        def is_required(name):
            info = self.registry.lookupElementInfo(name, info_dict)
            if info is None:
                return False
            if not info.required:
                self.logMsg('diag', 'keepOnlyRequired: element',
                            name, 'IS NOT required, skipping')
            return info.required

        return [name
                for name in names
                if is_required(name)]

    def makeReturnCodeList(self, attrib, cmd, name):
        """Return a list of possible return codes for a function.

        attrib is either 'successcodes' or 'errorcodes'.
        """
        return_lines = []
        RETURN_CODE_FORMAT = '* ename:{}'

        codes_attr = cmd.get(attrib)
        if codes_attr:
            codes = self.findRequiredEnums(codes_attr.split(','))
            if codes:
                return_lines.extend((RETURN_CODE_FORMAT.format(code)
                                     for code in codes))

        applicable_ext_codes = (ext_code
                                for ext_code in self.registry.commandextensionsuccesses
                                if ext_code.command == name)
        for ext_code in applicable_ext_codes:
            line = RETURN_CODE_FORMAT.format(ext_code.value)
            if ext_code.extension:
                line += ' [only if {} is enabled]'.format(
                    self.conventions.formatExtension(ext_code.extension))

            return_lines.append(line)
        if return_lines:
            return '\n'.join(return_lines)

        return None

    def makeSuccessCodes(self, cmd, name):
        return self.makeReturnCodeList('successcodes', cmd, name)

    def makeErrorCodes(self, cmd, name):
        return self.makeReturnCodeList('errorcodes', cmd, name)

    def recordStateRelationship(self, cmd, name, relationship: StateRelationship):
        states_str = cmd.get(STATE_ATTRIBUTES[relationship])
        if not states_str:
            return None

        state_names = states_str.split(',')
        for current_state in state_names:
            self.states.add_command(name, current_state, relationship)
        return state_names

    def genCmd(self, cmdinfo, name, alias):
        """Command generation."""
        OutputGenerator.genCmd(self, cmdinfo, name, alias)

        # @@@ (Jon) something needs to be done here to handle aliases, probably

        validity = self.makeValidityCollection(name)

        # OpenXR-only: make sure extension is enabled
        validity.possiblyAddExtensionRequirement(self.currentExtension, 'calling flink:')

        validity += self.makeStructOrCommandValidity(
            cmdinfo.elem, name, cmdinfo.getParams())

        threadsafety = self.makeThreadSafetyBlock(cmdinfo.elem, 'param')
        commandpropertiesentry = None

        # Vulkan-specific
        # commandpropertiesentry = self.makeCommandPropertiesTableEntry(
        #     cmdinfo.elem, name)
        successcodes = self.makeSuccessCodes(cmdinfo.elem, name)
        errorcodes = self.makeErrorCodes(cmdinfo.elem, name)

        # OpenXR-specific
        self.generateStateValidity(validity, name)

        self.writeInclude('protos', name, validity, threadsafety,
                          commandpropertiesentry, successcodes, errorcodes)

    def genStructInternals(self, typeinfo, typeName, validity, threadsafety):
        """Internals of generating for a struct, so aliases can be handled."""

        if typeinfo.elem.get('returnedonly') is None:
            validity += self.makeStructOrCommandValidity(
                typeinfo.elem, typeName, typeinfo.getMembers())
            threadsafety = self.makeThreadSafetyBlock(typeinfo.elem, 'member')

        else:
            # Need to generate structure type and next pointer chain member validation
            validity += self.makeOutputOnlyStructValidity(
                typeinfo.elem, typeName, typeinfo.getMembers())

    def genStruct(self, typeinfo, typeName, alias):
        """Struct Generation."""
        OutputGenerator.genStruct(self, typeinfo, typeName, alias)

        # @@@ (Jon) something needs to be done here to handle aliases, probably

        # Anything that's only ever returned cannot be set by the user, so should not have any validity information.
        validity = self.makeValidityCollection(typeName)
        threadsafety = []

        # OpenXR-only: make sure extension is enabled
        validity.possiblyAddExtensionRequirement(self.currentExtension, 'using slink:')

        self.genStructInternals(typeinfo, typeName, validity, threadsafety)

        if alias and self.conventions.duplicate_aliased_structs:
            validity.addValidityEntry(
                '**Note:** slink:{new} is an alias for slink:{old}, so implicit valid usage for slink:{old} has been replicated below.'.format(
                    old=alias, new=typeName))
            alias_info = self.registry.typedict[alias]
            self.genStructInternals(alias_info, alias, validity, threadsafety)

        self.writeInclude('structs', typeName, validity,
                          threadsafety, None, None, None)

    def genGroup(self, groupinfo, groupName, alias):
        """Group (e.g. C "enum" type) generation.
        For the validity generator, this just tags individual enumerants
        as required or not.
        """
        OutputGenerator.genGroup(self, groupinfo, groupName, alias)

        # @@@ (Jon) something needs to be done here to handle aliases, probably

        groupElem = groupinfo.elem

        # Loop over the nested 'enum' tags. Keep track of the minimum and
        # maximum numeric values, if they can be determined; but only for
        # core API enumerants, not extension enumerants. This is inferred
        # by looking for 'extends' attributes.
        for elem in groupElem.findall('enum'):
            name = elem.get('name')
            ei = self.registry.lookupElementInfo(name, self.registry.enumdict)

            # Tag enumerant as required or not
            ei.required = self.isEnumRequired(elem)

    def genType(self, typeinfo, name, alias):
        """Type Generation."""
        OutputGenerator.genType(self, typeinfo, name, alias)

        # @@@ (Jon) something needs to be done here to handle aliases, probably

        category = typeinfo.elem.get('category')
        if category in ('struct', 'union'):
            self.genStruct(typeinfo, name, alias)



#!/usr/bin/python3 -i
#
# Copyright (c) 2019-2024, The Khronos Group Inc.
# Copyright (c) 2019 Collabora, Ltd.
#
# SPDX-License-Identifier: Apache-2.0

"""Provides functionality to use Jinja2 when generating C/C++ code, while eliminating the need to import Jinja2 from any other file."""

import re
from pathlib import Path

_ADDED_TO_PATH = False


def _add_to_path():
    global _ADDED_TO_PATH
    if not _ADDED_TO_PATH:
        import sys
        # Find Jinja2 in source tree, as last resort.
        sys.path.append(
            str(
                Path(__file__).resolve().parent.parent.parent / "external" /
                "python"))
        _ADDED_TO_PATH = True


_WHITESPACE = re.compile(r"[\s\n]+")


def _undecorate(name):
    """Undecorate a name by removing the leading Xr and making it lowercase."""
    lower = name.lower()
    assert lower.startswith('xr')
    return lower[2:]


def _quote_string(s):
    return '"{}"'.format(s)


def _base_name(name):
    return name[2:]


def _collapse_whitespace(s):
    return _WHITESPACE.sub(" ", s)


def _protect_begin(entity):
    if entity.protect_value:
        return "#if {}".format(entity.protect_string)
    return ""


def _protect_end(entity):
    if entity.protect_value:
        return "#endif // {}".format(entity.protect_string)
    return ""

def _remove_prefix(s: str, prefix: str):
    if s.startswith(prefix):
        return s[len(prefix):]
    return s


def make_jinja_environment(file_with_templates_as_sibs=None, search_path=None):
    """Create a Jinja2 environment customized to generate C/C++ headers/code for Khronos APIs.

    Delimiters have been changed from Jinja2 defaults to permit better interoperability with
    editors and other tooling expecting C/C++ code, by combining them with comments:

    - Blocks are bracketed like /*% block_contents %*/ instead of {% block_contents %}
    - Variable outputs are bracketed like /*{ var }*/ instead of {{ var }}
    - Line statements start with //# instead of just #
    - Line comments start with //## instead of just ##

    Other details:

    - autoescape is turned off because this isn't HTML.
    - trailing newline kept
    - blocks are trimmed for easier markup.
    - the loader is a file system loader, building a search path from file_with_templates_as_sibs
      (if your template is a sibling of your source file, just pass file_with_templates_as_sibs=__file__),
      and search_path (an iterable if you want more control)

    Provided filters:

    - quote_string - wrap something in double-quotes
    - undecorate - same as the generator utility function: remove leading two-character API prefix and make lowercase.
    - base_name - just removes leading two-character API prefix
    - collapse_whitespace - minimizes internal whitespace with a regex

    Provided functions used as globals:

    - protect_begin(entity) and protect_end(entity) - use in a pair, and pass an entity
      (function/command, struct, etc), and if it is noted with protect="something" in the XML,
      the appropriate #if and #endif will be printed (for protect_begin and protect_end respectively.)

    You may further add globals and filters to the returned environment.
    """

    _add_to_path()
    from jinja2 import Environment, FileSystemLoader

    search_paths = []
    if file_with_templates_as_sibs:
        search_paths.append(
            str(Path(file_with_templates_as_sibs).parent))
    if search_path:
        search_paths.extend(search_path)
    env = Environment(keep_trailing_newline=True,
                      trim_blocks=True,
                      block_start_string="/*%",
                      block_end_string="%*/",
                      variable_start_string="/*{",
                      variable_end_string="}*/",
                      line_statement_prefix="//#",
                      line_comment_prefix="//##",
                      autoescape=False,
                      loader=FileSystemLoader(search_paths))
    env.filters['quote_string'] = _quote_string
    env.filters['undecorate'] = _undecorate
    env.filters['base_name'] = _base_name
    env.filters['collapse_whitespace'] = _collapse_whitespace
    env.filters['remove_prefix'] = _remove_prefix
    env.globals['protect_begin'] = _protect_begin
    env.globals['protect_end'] = _protect_end

    return env


class JinjaTemplate:
    def __init__(self, env, fn):
        """Load and parse a Jinja2 template given a Jinja2 environment and the template file name.

        Create the environment using make_jinja_environment().

        Syntax errors are caught, have their details printed, then are re-raised (to stop execution).
        """
        _add_to_path()
        from jinja2 import TemplateSyntaxError
        try:
            self.template = env.get_template(fn)
        except TemplateSyntaxError as e:
            print("Jinja2 template syntax error during parse: {}:{} error: {}".
                  format(e.filename, e.lineno, e.message))
            raise e

    def render(self, *args, **kwargs):
        """Render the Jinja2 template with the provided context.

        All arguments are passed through; this just wraps the Jinja2 template render method
        to handle syntax error exceptions so that Jinja2 does not need to be imported anywhere
        but this file.
        """
        _add_to_path()
        from jinja2 import TemplateSyntaxError

        try:
            return self.template.render(*args, **kwargs)
        except TemplateSyntaxError as e:
            print(
                "Jinja2 template syntax error during render: {}:{} error: {}".
                format(e.filename, e.lineno, e.message))
            raise e



#!/usr/bin/python3
#
# Copyright (c) 2017-2024, The Khronos Group Inc.
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import getopt
import os
import sys

cur_layer_json_version = '1.0.0'

# Get the relative path of a JSON's library file using the JSON
# file itself as a starting point.
#   json_file       the JSON file we're creating
#   library_file    the library (so/dll) file we're referencing in the JSON file


def getRelativePath(json_file, library_file):
    directory_slash = '/'
    relative_path = ''
    relative_list = []
    up_folder_count = 0
    json_path = os.path.normpath(json_file)
    json_split = json_path.split(os.sep)
    json_len = len(json_split)
    lib_path = os.path.normpath(library_file)
    lib_split = lib_path.split(os.sep)
    lib_len = len(lib_split)

    smallest_count = json_len - 1
    if lib_len < smallest_count:
        smallest_count = lib_len - 1

    cur_folder = 0
    diff_base_found = False
    for i in range(smallest_count):
        if json_split[i] != lib_split[i]:
            diff_base_found = True

        if diff_base_found:
            up_folder_count = up_folder_count + 1
            relative_list.append(lib_split[i])

        cur_folder = cur_folder + 1

    if json_len - 1 > smallest_count:
        up_folder_count += json_len - smallest_count - 1

    for i in range(smallest_count, lib_len):
        relative_list.append(lib_split[i])

    if up_folder_count == 0:
        relative_path += '.'
        relative_path += directory_slash
    else:
        for i in range(up_folder_count):
            relative_path += '..'
            relative_path += directory_slash

    for i in range(len(relative_list)):
        if i > 0:
            relative_path += directory_slash
        relative_path += relative_list[i]

    print(relative_path)
    return relative_path


def main(argv):
    output_file = ''
    layer_name = ''
    library_location = ''
    api_version = ''
    implementation_version = ''
    description = ''
    generate_badjson_jsons = False

    usage = '\ngenerate_api_layer_manifest.py <ARGS>\n'
    usage += '    -f/--file <filename>\n'
    usage += '    -n/--name <layer name>\n'
    usage += '    -l/--lib <library_location>\n'
    usage += '    -a/--api <OpenXR API version>\n'
    usage += '    -v/--ver <layer implementation version>\n'
    usage += '    -d/--desc <Description>\n'
    usage += '    -b/--bad\n'

    try:
        opts, _ = getopt.getopt(argv, "hbf:n:l:a:v:d:", ["bad", "file=", "name=", "lib=", "api=", "ver=", "desc="])
    except getopt.GetoptError:
        print(usage)
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print(usage)
            sys.exit()
        elif opt in ("-f", "--file"):
            output_file = arg.strip()
        elif opt in ("-n", "--name"):
            layer_name = arg.strip()
        elif opt in ("-l", "--lib"):
            library_location = arg.strip()
        elif opt in ("-a", "--api"):
            api_version = arg.strip()
        elif opt in ("-v", "--ver"):
            implementation_version = arg.strip()
        elif opt in ("-d", "--desc"):
            description = arg.strip()
        elif opt in ("-b", "--bad"):
            generate_badjson_jsons = True

    file_text = '{\n'
    file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
    file_text += '    "api_layer": {\n'
    file_text += '        "name": "XR_APILAYER_%s",\n' % layer_name
    file_text += '        "library_path": "%s",\n' % library_location
    file_text += '        "api_version": "%s",\n' % api_version
    file_text += '        "implementation_version": "%s",\n' % implementation_version
    file_text += '        "description": "%s"' % description

    # If testing bad JSONs, then add in a fake extension
    if generate_badjson_jsons:
        file_text += ',\n'
        file_text += '        "instance_extensions": [\n'
        file_text += '            {\n'
        file_text += '                "name": "XR_KHR_fake_ext2",\n'
        file_text += '                "extension_version": "12"\n'
        file_text += '            },\n'
        file_text += '            {\n'
        file_text += '                "name": "XR_KHR_fake_ext3",\n'
        file_text += '                "extension_version": "42"\n'
        file_text += '            }\n'
        file_text += '        ]'
    file_text += '\n'
    file_text += '    }\n'
    file_text += '}\n'
    f = open(output_file, 'w')
    f.write(file_text)
    f.close()

    if generate_badjson_jsons:
        # Bad File format versions
        ####################################

        # Missing
        bad_name = '_badjson_file_ver_missing'
        file_text = '{\n'
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Use int
        bad_name = '_badjson_file_ver_int'
        file_text = '{\n'
        file_text += '    "file_format_version": 1,\n'
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Use invalid string
        bad_name = '_badjson_file_ver_string'
        file_text = '{\n'
        file_text += '    "file_format_version": "invalid string",\n'
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Too low of a version
        bad_name = '_badjson_file_ver_all_low'
        file_text = '{\n'
        file_text += '    "file_format_version": "0.0.0",\n'
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Too high of a major version
        bad_name = '_badjson_file_ver_major_high'
        file_text = '{\n'
        file_text += '    "file_format_version": "15.0.0",\n'
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Too high of a minor version
        bad_name = '_badjson_file_ver_minor_high'
        file_text = '{\n'
        file_text += '    "file_format_version": "1.15.0",\n'
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # No layer section
        ####################################

        # Completely Missing
        bad_name = '_badjson_layer_missing'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '    "library_path": "%s",\n' % library_location
        file_text += '    "api_version": "%s",\n' % api_version
        file_text += '    "implementation_version": "%s",\n' % implementation_version
        file_text += '    "description": "%s"\n' % description
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Empty
        bad_name = '_badjson_layer_empty'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '    },\n'
        file_text += '    "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '    "library_path": "%s",\n' % library_location
        file_text += '    "api_version": "%s",\n' % api_version
        file_text += '    "implementation_version": "%s",\n' % implementation_version
        file_text += '    "description": "%s"\n' % description
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Bad Layer Name
        ####################################

        # Use int
        bad_name = '_badjson_name_int'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '        "name": 1,\n'
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Missing
        bad_name = '_badjson_name_missing'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Bad path
        ####################################

        # Missing
        bad_name = '_badjson_path_missing'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Use int
        bad_name = '_badjson_path_int'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": 1,\n'
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Replace valid path with invalid one
        bad_name = '_badjson_path_no_file'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": "%s",\n' % library_location.replace("test_layers", "not_real")
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Bad API Version
        ####################################

        # Missing
        bad_name = '_badjson_api_int'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Use int
        bad_name = '_badjson_api_int'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "api_version": 1,\n'
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Use float
        bad_name = '_badjson_api_float'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "api_version": 1.0,\n'
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Bad string
        bad_name = '_badjson_api_string'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "api_version": "string",\n'
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Too high of a major API version
        bad_name = '_badjson_api_major_high'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "api_version": "15.0",\n'
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Valid JSON, but invalid Negotiate
        ####################################

        # Always fail negotiate
        bad_name = '_badnegotiate_always'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s",\n' % description
        file_text += '        "functions": {\n'
        file_text += '           "xrNegotiateLoaderApiLayerInterface":\n'
        file_text += '               "TestLayerAlwaysFailNegotiateLoaderApiLayerInterface"\n'
        file_text += '       }\n'
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Pass negotiate, but return null GIPA
        bad_name = '_badnegotiate_invalid_gipa'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s",\n' % description
        file_text += '        "functions": {\n'
        file_text += '           "xrNegotiateLoaderApiLayerInterface":\n'
        file_text += '               "TestLayerNullGipaNegotiateLoaderApiLayerInterface"\n'
        file_text += '       }\n'
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Pass negotiate, but return invalid interface version
        bad_name = '_badnegotiate_invalid_interface'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s",\n' % description
        file_text += '        "functions": {\n'
        file_text += '           "xrNegotiateLoaderApiLayerInterface":\n'
        file_text += '               "TestLayerInvalidInterfaceNegotiateLoaderApiLayerInterface"\n'
        file_text += '       }\n'
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Pass negotiate, but return invalid api version
        bad_name = '_badnegotiate_invalid_api'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, bad_name)
        file_text += '        "library_path": "%s",\n' % library_location
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s",\n' % description
        file_text += '        "functions": {\n'
        file_text += '           "xrNegotiateLoaderApiLayerInterface":\n'
        file_text += '               "TestLayerInvalidApiNegotiateLoaderApiLayerInterface"\n'
        file_text += '       }\n'
        file_text += '    }\n'
        file_text += '}\n'
        bad_name += '.json'
        bad_file = output_file.replace(".json", bad_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Valid JSON, with relative path to library
        ####################################
        relative_lib = getRelativePath(output_file, library_location)

        # Provide a good relative path
        layer_suffix_name = '_good_relative_path'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, layer_suffix_name)
        file_text += '        "library_path": "%s",\n' % relative_lib
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        layer_suffix_name += '.json'
        bad_file = output_file.replace(".json", layer_suffix_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()

        # Provide a bad relative path
        layer_suffix_name = '_badjson_relative_path'
        file_text = '{\n'
        file_text += '    "file_format_version": "%s",\n' % cur_layer_json_version
        file_text += '    "api_layer": {\n'
        file_text += '        "name": "XR_APILAYER_LUNARG_%s%s",\n' % (layer_name, layer_suffix_name)
        file_text += '        "library_path": "%s",\n' % relative_lib.replace("test_layers", "not_real")
        file_text += '        "api_version": "%s",\n' % api_version
        file_text += '        "implementation_version": "%s",\n' % implementation_version
        file_text += '        "description": "%s"\n' % description
        file_text += '    }\n'
        file_text += '}\n'
        layer_suffix_name += '.json'
        bad_file = output_file.replace(".json", layer_suffix_name)
        f = open(bad_file, 'w')
        f.write(file_text)
        f.close()


if __name__ == "__main__":
    main(sys.argv[1:])



"""Provides a re-usable command-line interface to a MacroChecker."""

# Copyright (c) 2018-2019 Collabora, Ltd.
#
# SPDX-License-Identifier: Apache-2.0
#
# Author(s):    Rylie Pavlik <rylie.pavlik@collabora.com>


import argparse
import logging
import re
from pathlib import Path

from .shared import MessageId


def checkerMain(default_enabled_messages, make_macro_checker,
                all_docs, available_messages=None):
    """Perform the bulk of the work for a command-line interface to a MacroChecker.

    Arguments:
    default_enabled_messages -- The MessageId values that should be enabled by default.
    make_macro_checker -- A function that can be called with a set of enabled MessageId to create a
      properly-configured MacroChecker.
    all_docs -- A list of all spec documentation files.
    available_messages -- a list of all MessageId values that can be generated for this project.
      Defaults to every value. (e.g. some projects don't have MessageId.LEGACY)
    """
    enabled_messages = set(default_enabled_messages)
    if not available_messages:
        available_messages = list(MessageId)

    disable_args = []
    enable_args = []

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--scriptlocation",
        help="Append the script location generated a message to the output.",
        action="store_true")
    parser.add_argument(
        "--verbose",
        "-v",
        help="Output 'info'-level development logging messages.",
        action="store_true")
    parser.add_argument(
        "--debug",
        "-d",
        help="Output 'debug'-level development logging messages (more verbose than -v).",
        action="store_true")
    parser.add_argument(
        "-Werror",
        "--warning_error",
        help="Make warnings act as errors, exiting with non-zero error code",
        action="store_true")
    parser.add_argument(
        "--include_warn",
        help="List all expected but unseen include files, not just those that are referenced.",
        action='store_true')
    parser.add_argument(
        "-Wmissing_refpages",
        help="List all entities with expected but unseen ref page blocks. NOT included in -Wall!",
        action='store_true')
    parser.add_argument(
        "--include_error",
        help="Make expected but unseen include files cause exiting with non-zero error code",
        action='store_true')
    parser.add_argument(
        "--broken_error",
        help="Make missing include/anchor for linked-to entities cause exiting with non-zero error code. Weaker version of --include_error.",
        action='store_true')
    parser.add_argument(
        "--dump_entities",
        help="Just dump the parsed entity data to entities.json and exit.",
        action='store_true')
    parser.add_argument(
        "--html",
        help="Output messages to the named HTML file instead of stdout.")
    parser.add_argument(
        "file",
        help="Only check the indicated file(s). By default, all chapters and extensions are checked.",
        nargs="*")
    parser.add_argument(
        "--ignore_count",
        type=int,
        help="Ignore up to the given number of errors without exiting with a non-zero error code.")
    parser.add_argument("-Wall",
                        help="Enable all warning categories.",
                        action='store_true')

    for message_id in MessageId:
        enable_arg = message_id.enable_arg()
        enable_args.append((message_id, enable_arg))

        disable_arg = message_id.disable_arg()
        disable_args.append((message_id, disable_arg))
        if message_id in enabled_messages:
            parser.add_argument('-' + disable_arg, action="store_true",
                                help="Disable message category {}: {}".format(str(message_id), message_id.desc()))
            # Don't show the enable flag in help since it's enabled by default
            parser.add_argument('-' + enable_arg, action="store_true",
                                help=argparse.SUPPRESS)
        else:
            parser.add_argument('-' + enable_arg, action="store_true",
                                help="Enable message category {}: {}".format(str(message_id), message_id.desc()))
            # Don't show the disable flag in help since it's disabled by
            # default
            parser.add_argument('-' + disable_arg, action="store_true",
                                help=argparse.SUPPRESS)

    args = parser.parse_args()

    arg_dict = vars(args)
    for message_id, arg in enable_args:
        if args.Wall or (arg in arg_dict and arg_dict[arg]):
            enabled_messages.add(message_id)

    for message_id, arg in disable_args:
        if arg in arg_dict and arg_dict[arg]:
            enabled_messages.discard(message_id)

    if args.verbose:
        logging.basicConfig(level='INFO')

    if args.debug:
        logging.basicConfig(level='DEBUG')

    checker = make_macro_checker(enabled_messages)

    if args.dump_entities:
        with open('entities.json', 'w', encoding='utf-8') as f:
            f.write(checker.getEntityJson())
            exit(0)

    if args.file:
        files = (str(Path(f).resolve()) for f in args.file)
    else:
        files = all_docs

    for fn in files:
        checker.processFile(fn)

    if args.html:
        from .html_printer import HTMLPrinter
        printer = HTMLPrinter(args.html)
    else:
        from .console_printer import ConsolePrinter
        printer = ConsolePrinter()

    if args.scriptlocation:
        printer.show_script_location = True

    if args.file:
        printer.output("Only checked specified files.")
        for f in args.file:
            printer.output(f)
    else:
        printer.output("Checked all chapters and extensions.")

    if args.warning_error:
        numErrors = checker.numDiagnostics()
    else:
        numErrors = checker.numErrors()

    check_includes = args.include_warn
    check_broken = not args.file

    if args.file and check_includes:
        print('Note: forcing --include_warn off because only checking supplied files.')
        check_includes = False

    printer.outputResults(checker, broken_links=(not args.file),
                          missing_includes=check_includes)

    if check_broken:
        numErrors += len(checker.getBrokenLinks())

    if args.file and args.include_error:
        print('Note: forcing --include_error off because only checking supplied files.')
        args.include_error = False
    if args.include_error:
        numErrors += len(checker.getMissingUnreferencedApiIncludes())

    check_missing_refpages = args.Wmissing_refpages
    if args.file and check_missing_refpages:
        print('Note: forcing -Wmissing_refpages off because only checking supplied files.')
        check_missing_refpages = False

    if check_missing_refpages:
        missing = checker.getMissingRefPages()
        if missing:
            printer.output("Expected, but did not find, ref page blocks for the following {} entities: {}".format(
                len(missing),
                ', '.join(missing)
            ))
            if args.warning_error:
                numErrors += len(missing)

    printer.close()

    if args.broken_error and not args.file:
        numErrors += len(checker.getBrokenLinks())

    if checker.hasFixes():
        fixFn = 'applyfixes.sh'
        print('Saving shell script to apply fixes as {}'.format(fixFn))
        with open(fixFn, 'w', encoding='utf-8') as f:
            f.write('#!/bin/sh -e\n')
            for fileChecker in checker.files:
                wroteComment = False
                for msg in fileChecker.messages:
                    if msg.fix is not None:
                        if not wroteComment:
                            f.write('\n# {}\n'.format(fileChecker.filename))
                            wroteComment = True
                        search, replace = msg.fix
                        f.write(
                            r"sed -i -r 's~\b{}\b~{}~g' {}".format(
                                re.escape(search),
                                replace,
                                fileChecker.filename))
                        f.write('\n')

    print('Total number of errors with this run: {}'.format(numErrors))

    if args.ignore_count:
        if numErrors > args.ignore_count:
            # Exit with non-zero error code so that we "fail" CI, etc.
            print('Exceeded specified limit of {}, so exiting with error'.format(
                args.ignore_count))
            exit(1)
        else:
            print('At or below specified limit of {}, so exiting with success'.format(
                args.ignore_count))
            exit(0)

    if numErrors:
        # Exit with non-zero error code so that we "fail" CI, etc.
        print('Exiting with error')
        exit(1)
    else:
        print('Exiting with success')
        exit(0)



"""Defines ConsolePrinter, a BasePrinter subclass for appealing console output."""

# Copyright (c) 2018-2019 Collabora, Ltd.
#
# SPDX-License-Identifier: Apache-2.0
#
# Author(s):    Rylie Pavlik <rylie.pavlik@collabora.com>

from sys import stdout

from .base_printer import BasePrinter
from .shared import (colored, getHighlightedRange, getInterestedRange,
                     toNameAndLine)

try:
    from tabulate import tabulate_impl
    HAVE_TABULATE = True
except ImportError:
    HAVE_TABULATE = False


def colWidth(collection, columnNum):
    """Compute the required width of a column in a collection of row-tuples."""
    MIN_PADDING = 5
    return MIN_PADDING + max((len(row[columnNum]) for row in collection))


def alternateTabulate(collection, headers=None):
    """Minimal re-implementation of the tabulate module."""
    # We need a list, not a generator or anything else.
    if not isinstance(collection, list):
        collection = list(collection)

    # Empty collection means no table
    if not collection:
        return None

    if headers is None:
        fullTable = collection
    else:
        underline = ['-' * len(header) for header in headers]
        fullTable = [headers, underline] + collection
    widths = [colWidth(collection, colNum)
              for colNum in range(len(fullTable[0]))]
    widths[-1] = None

    lines = []
    for row in fullTable:
        fields = []
        for data, width in zip(row, widths):
            if width:
                spaces = ' ' * (width - len(data))
                fields.append(data + spaces)
            else:
                fields.append(data)
        lines.append(''.join(fields))
    return '\n'.join(lines)


def printTabulated(collection, headers=None):
    """Call either tabulate.tabulate(), or our internal alternateTabulate()."""
    if HAVE_TABULATE:
        tabulated = tabulate_impl(collection, headers=headers)
    else:
        tabulated = alternateTabulate(collection, headers=headers)
    if tabulated:
        print(tabulated)


def printLineSubsetWithHighlighting(
        line, start, end, highlightStart=None, highlightEnd=None, maxLen=120, replacement=None):
    """Print a (potential subset of a) line, with highlighting/underline and optional replacement.

    Will print at least the characters line[start:end], and potentially more if possible
    to do so without making the output too wide.
    Will highlight (underline) line[highlightStart:highlightEnd], where the default
    value for highlightStart is simply start, and the default value for highlightEnd is simply end.
    replacement, if supplied, will be aligned with the highlighted range.

    Output is intended to look like part of a Clang compile error/warning message.
    """
    # Fill in missing start/end with start/end of range.
    if highlightStart is None:
        highlightStart = start
    if highlightEnd is None:
        highlightEnd = end

    # Expand interested range start/end.
    start = min(start, highlightStart)
    end = max(end, highlightEnd)

    tildeLength = highlightEnd - highlightStart - 1
    caretLoc = highlightStart
    continuation = '[...]'

    if len(line) > maxLen:
        # Too long

        # the max is to handle -1 from .find() (which indicates "not found")
        followingSpaceIndex = max(end, line.find(' ', min(len(line), end + 1)))

        # Maximum length has decreased by at least
        # the length of a single continuation we absolutely need.
        maxLen -= len(continuation)

        if followingSpaceIndex <= maxLen:
            # We can grab the whole beginning of the line,
            # and not adjust caretLoc
            line = line[:maxLen] + continuation

        elif (len(line) - followingSpaceIndex) < 5:
            # We need to truncate the beginning,
            # but we're close to the end of line.
            newBeginning = len(line) - maxLen

            caretLoc += len(continuation)
            caretLoc -= newBeginning
            line = continuation + line[newBeginning:]
        else:
            # Need to truncate the beginning of the string too.
            newEnd = followingSpaceIndex

            # Now we need two continuations
            # (and to adjust caret to the right accordingly)
            maxLen -= len(continuation)
            caretLoc += len(continuation)

            newBeginning = newEnd - maxLen
            caretLoc -= newBeginning

            line = continuation + line[newBeginning:newEnd] + continuation

    stdout.buffer.write(line.encode('utf-8'))
    print()

    spaces = ' ' * caretLoc
    tildes = '~' * tildeLength
    print(spaces + colored('^' + tildes, 'green'))
    if replacement is not None:
        print(spaces + colored(replacement, 'green'))


class ConsolePrinter(BasePrinter):
    """Implementation of BasePrinter for generating diagnostic reports in colored, helpful console output."""

    def __init__(self):
        self.show_script_location = False
        super().__init__()

    ###
    # Output methods: these all print directly.
    def outputResults(self, checker, broken_links=True,
                      missing_includes=False):
        """Output the full results of a checker run.

        Includes the diagnostics, broken links (if desired),
        and missing includes (if desired).
        """
        self.output(checker)
        if broken_links:
            broken = checker.getBrokenLinks()
            if broken:
                self.outputBrokenLinks(checker, broken)
        if missing_includes:
            missing = checker.getMissingUnreferencedApiIncludes()
            if missing:
                self.outputMissingIncludes(checker, missing)

    def outputBrokenLinks(self, checker, broken):
        """Output a table of broken links.

        Called by self.outputBrokenAndMissing() if requested.
        """
        print('Missing API includes that are referenced by a linking macro: these result in broken links in the spec!')

        def makeRowOfBroken(entity, uses):
            fn = checker.findEntity(entity).filename
            anchor = '[[{}]]'.format(entity)
            locations = ', '.join((toNameAndLine(context, root_path=checker.root_path)
                                   for context in uses))
            return (fn, anchor, locations)
        printTabulated((makeRowOfBroken(entity, uses)
                        for entity, uses in sorted(broken.items())),
                       headers=['Include File', 'Anchor in lieu of include', 'Links to this entity'])

    def outputMissingIncludes(self, checker, missing):
        """Output a table of missing includes.

        Called by self.outputBrokenAndMissing() if requested.
        """
        missing = list(sorted(missing))
        if not missing:
            # Exit if none
            return
        print(
            'Missing, but unreferenced, API includes/anchors - potentially not-documented entities:')

        def makeRowOfMissing(entity):
            fn = checker.findEntity(entity).filename
            anchor = '[[{}]]'.format(entity)
            return (fn, anchor)
        printTabulated((makeRowOfMissing(entity) for entity in missing),
                       headers=['Include File', 'Anchor in lieu of include'])

    def outputMessage(self, msg):
        """Output a Message, with highlighted range and replacement, if appropriate."""
        highlightStart, highlightEnd = getHighlightedRange(msg.context)

        if '\n' in msg.context.filename:
            # This is a multi-line string "filename".
            # Extra blank line and delimiter line for readability:
            print()
            print('--------------------------------------------------------------------')

        fileAndLine = colored('{}:'.format(
            self.formatBrief(msg.context)), attrs=['bold'])

        headingSize = len('{context}: {mtype}: '.format(
            context=self.formatBrief(msg.context),
            mtype=self.formatBrief(msg.message_type, False)))
        indent = ' ' * headingSize
        printedHeading = False

        lines = msg.message[:]
        if msg.see_also:
            lines.append('See also:')
            lines.extend(('  {}'.format(self.formatBrief(see))
                          for see in msg.see_also))

        if msg.fix:
            lines.append('Note: Auto-fix available')

        for line in msg.message:
            if not printedHeading:
                scriptloc = ''
                if msg.script_location and self.show_script_location:
                    scriptloc = ', ' + msg.script_location
                print('{fileLine} {mtype} {msg} (-{arg}{loc})'.format(
                    fileLine=fileAndLine, mtype=msg.message_type.formattedWithColon(),
                    msg=colored(line, attrs=['bold']), arg=msg.message_id.enable_arg(), loc=scriptloc))
                printedHeading = True
            else:
                print(colored(indent + line, attrs=['bold']))

        if len(msg.message) > 1:
            # extra blank line after multiline message
            print('')

        start, end = getInterestedRange(msg.context)
        printLineSubsetWithHighlighting(
            msg.context.line,
            start, end,
            highlightStart, highlightEnd,
            replacement=msg.replacement)

    def outputFallback(self, obj):
        """Output by calling print."""
        print(obj)

    ###
    # Format methods: these all return a string.
    def formatFilename(self, fn, _with_color=True):
        """Format a local filename, as a relative path if possible."""
        return self.getRelativeFilename(fn)

    def formatMessageTypeBrief(self, message_type, with_color=True):
        """Format a message type briefly, applying color if desired and possible.

        Delegates to the superclass if not formatting with color.
        """
        if with_color:
            return message_type.formattedWithColon()
        return super(ConsolePrinter, self).formatMessageTypeBrief(
            message_type, with_color)



#!/usr/bin/python3 -i
#
# Copyright (c) 2013-2024, The Khronos Group Inc.
#
# SPDX-License-Identifier: Apache-2.0

# Working-group-specific style conventions,
# used in generation.

import re
from pathlib import Path

from spec_tools.conventions import ConventionsBase

# Tokenize into "words"
# The leading named groups are for "special cases" per spec "Implicit Valid Usage" section 2.7.7
# Note that while Vulkan is listed as a special case,
# it doesn't actually break the rules, so it's not treated specially
MAIN_RE = re.compile(
    r'''(?P<d3d>D3D[0-9]*)|    # Take D3D as a group of its own
        (?P<egl>EGL)|          # EGL as a word of its own
        (?P<gl>OpenGL(ES)?)|   # OpenGL and OpenGLES as words
        (?P<dimension>[0-9]D)| # Things like 2D are words
        (?P<word>              # Normal-ish words, which are....
            ([A-Z]([a-z]+([0-9](?!D[A-Z]{1}))*)+)|  # Capital letter followed by at least one lowercase letter, possibly ending in some digits as long as the digits aren't followed by a "D" then another word
            ([A-Z][A-Z0-9]+(?![a-z]))       # Or, all-caps letter and digit mix starting with a letter, excluding the last capital before some lowercase
        )''', re.VERBOSE)

class OpenXRConventions(ConventionsBase):
    """The specifics of how OpenXR writes a spec."""

    @property
    def null(self):
        """Preferred spelling of NULL."""
        return 'code:NULL'

    @property
    def constFlagBits(self):
        """Return True if static const flag bits should be generated, False if an enumerated type should be generated."""
        return True

    @property
    def struct_macro(self):
        """Get the appropriate format macro for a structure.

        Primarily affects generated valid usage statements.
        """

        return 'slink:'

    @property
    def structtype_member_name(self):
        """Return name of the structure type member"""
        return 'type'

    @property
    def nextpointer_member_name(self):
        """Return name of the structure pointer chain member"""
        return 'next'

    @property
    def valid_pointer_prefix(self):
        """Return prefix to pointers which must themselves be valid.

           OpenXR doesn't currently do this.
        """
        return None

    def is_structure_type_member(self, paramtype, paramname):
        """Determine if member type and name match the structure type member."""
        return paramtype == 'XrStructureType' and paramname == self.structtype_member_name

    def is_nextpointer_member(self, paramtype, paramname):
        """Determine if member type and name match the next pointer chain member."""
        return paramtype == 'void' and paramname == self.nextpointer_member_name

    def generate_structure_type_from_name(self, structname):
        """Generate a structure type name, like XR_TYPE_CREATE_INSTANCE_INFO"""
        structure_type_parts = []
        # Tokenize into "words"
        for elem in MAIN_RE.finditer(structname):
            if elem.group('d3d'):
                # D3D ⇒ _D3D
                structure_type_parts.append(elem.group())
            elif elem.group('egl'):
                # EGL ⇒ _EGL
                structure_type_parts.append(elem.group())
            elif elem.group('gl'):
                # OpenGL ⇒ _OPENGL
                # OpenGLES ⇒ _OPENGL_ES
                structure_type_parts.append(
                    elem.group().upper().replace('ES', '_ES'))
            elif elem.group('dimension'):
                # 2D ⇒ _2D
                structure_type_parts.append(
                    elem.group())
            else:
                word = elem.group('word')
                if word == 'Xr':
                    structure_type_parts.append('XR_TYPE')
                else:
                    structure_type_parts.append(word.upper())
        return '_'.join(structure_type_parts)

    @property
    def warning_comment(self):
        """Return warning comment to be placed in header of generated Asciidoctor files"""
        return '// WARNING: DO NOT MODIFY! This file is automatically generated from the xr.xml registry'

    @property
    def file_suffix(self):
        """Return suffix of generated Asciidoctor files"""
        return '.adoc'

    def api_name(self, spectype='api'):
        """Return API or specification name for citations in ref pages.

        spectype is the spec this refpage is for.
        'api' (the default value) is the main OpenXR API Specification.
        If an unrecognized spectype is given, returns None.

        """
        if spectype == 'api' or spectype is None:
            return 'OpenXR'
        return None

    @property
    def xml_supported_name_of_api(self):
        """Return the supported= attribute used in API XML"""
        return 'openxr'

    @property
    def api_prefix(self):
        """Return API token prefix"""
        return 'XR_'

    @property
    def write_contacts(self):
        """Return whether contact list should be written to extension appendices"""
        return False

    @property
    def write_refpage_include(self):
        """Return whether refpage include should be written to extension appendices"""
        return True

    def writeFeature(self, featureName, featureExtraProtect, filename):
        """Returns True if OutputGenerator.endFeature should write this feature.

        Used in COutputGenerator.

        In OpenXR, a feature is written to the openxr_platform header
        if and only if it has a defined "protect" attribute.
        """
        if filename == 'openxr_reflection.h':
            # Write all features to the reflection header
            return True
        is_loader = featureName == 'XR_LOADER_VERSION_1_0'
        is_protected = featureExtraProtect is not None
        is_platform_header = (filename == 'openxr_platform.h')

        # Only write loader spec to loader file.
        if is_loader:
            return filename == 'openxr_loader_negotiation.h'

        if filename == 'openxr_loader_negotiation.h':
            return False

        # non-protected goes in non-platform header,
        # protected goes in platform header.
        return is_protected == is_platform_header

    def requires_error_validation(self, return_type):
        """Returns True if the return_type element is an API result code
           requiring error validation.
        """
        return return_type is not None and return_type.text == 'XrResult'

    @property
    def member_used_for_unique_vuid(self):
        """Return the member name used in the VUID-...-...-unique ID."""
        return self.nextpointer_member_name

    def is_externsync_command(self, protoname):
        """Returns True if the protoname element is an API command requiring
           external synchronization.

           OpenXR does not do this at present.
        """
        return False

    def is_api_name(self, name):
        """Returns True if name is in the reserved API namespace.
        For OpenXR, these are names with a case-insensitive 'xr' prefix, or
        a 'PFN_xr' function pointer type prefix.
        """
        return name[0:2].lower() == 'xr' or name[0:6] == 'PFN_xr'

    def should_insert_may_alias_macro(self, genOpts):
        """Return true if we should insert a "may alias" macro in this file."""

        # Don't insert may alias macro in things included in the spec.
        return "inc" not in genOpts.filename

    def make_voidpointer_alias(self, tail):
        """Reformat a void * declaration to include the API alias macro"""
        return '* XR_MAY_ALIAS{}'.format(tail[1:])

    def specURL(self, spectype='api'):
        """Return public registry URL which ref pages should link to for the
           current all-extensions HTML specification, so xrefs in the
           asciidoc source that aren't to ref pages can link into it
           instead. N.b. this may need to change on a per-refpage basis if
           there are multiple documents involved.
        """
        return 'https://registry.khronos.org/OpenXR/specs/1.0/html/xrspec.html'

    @property
    def xml_api_name(self):
        """Return the name used in the default API XML registry for the default API"""
        return 'openxr'

    @property
    def registry_path(self):
        """Return relpath to the default API XML registry in this project."""
        return 'registry/xr.xml'

    @property
    def specification_path(self):
        """Return relpath to the Asciidoctor specification sources in this project."""
        return '{appendices}'

    @property
    def extra_refpage_headers(self):
        """Return any extra text to add to refpage headers."""
        return 'include::{config}/attribs.adoc[]'

    @property
    def extension_index_prefixes(self):
        """Return a set of extension prefixes used to group extension refpages."""
        return ('XR_KHR', 'XR_KHX', 'XR_EXT', 'XR')

    @property
    def unified_flag_refpages(self):
        """Return True if Flags/FlagBits refpages are unified, False if
           they're separate.
        """
        return False

    @property
    def spec_reflow_path(self):
        """Return the path to the spec source folder to reflow"""
        return str(Path(__file__).resolve().parent.parent / 'sources')

    @property
    def spec_no_reflow_dirs(self):
        """Return a set of directories not to automatically descend into
           when reflowing spec text
        """
        return tuple()

    @property
    def zero(self):
        """Return the preferred way of formatting a literal 0."""
        return "`0`"

    @property
    def generate_index_terms(self):
        """Return True if asiidoctor index terms should be generated as part
           of an API interface from the docgenerator."""

        return True

    @property
    def generate_enum_table(self):
        """Return True if asciidoctor tables describing enumerants in a
           group should be generated as part of group generation."""
        return True

    def generate_max_enum_in_docs(self):
        """Return True if MAX_ENUM tokens should be generated in
           documentation includes."""
        return False

    def extension_file_path(self, name):
        """Return file path to an extension appendix relative to a directory
           containing all such appendices.
           - name - extension name"""

        (vendor, bare_name) = self.extension_name_split(name)
        vendor = vendor.lower()
        bare_name = bare_name.lower()

        return '{vendor}/{vendor}_{bare_name}{suffix}'.format(
                vendor=vendor,
                bare_name=bare_name,
                suffix=self.file_suffix)

    def extension_include_string(self, name):
        """Return format string for include:: line for an extension appendix
           file.
            - name - extension name"""

        return 'include::{{appendices}}/{}[]'.format(
                self.extension_file_path(name))

    @property
    def provisional_extension_warning(self):
        """Return True if a warning should be included in extension
           appendices for provisional extensions."""
        return False

    @property
    def include_extension_appendix_in_refpage(self):
        """Return True if generating extension refpages by embedding
           extension appendix content (default), False otherwise
           (OpenXR)."""

        return False

    @property
    def duplicate_aliased_structs(self):
        """
        Should aliased structs have the original struct definition listed in the
        generated docs snippet?
        """
        return True

    @property
    def protectProtoComment(self):
        """Return True if generated #endif should have a comment matching
           the protection symbol used in the opening #ifdef/#ifndef."""
        return True
